import inspect
import typing

from di.typing import get_markers_from_parameter
from starlette.datastructures import FormData

from xpresso._utils.typing import Some, model_field_from_param
from xpresso.binders._body.form_field import FormDataExtractor
from xpresso.binders.api import BodyExtractor
from xpresso.binders.dependants import BodyBinderMarker


class FieldExtractorBase:
    __slots__ = ("field_name", "field_extractor")

    def __init__(
        self,
        field_name: str,
        field_extractor: BodyExtractor,
    ) -> None:
        self.field_name = field_name
        self.field_extractor = field_extractor


class FieldExtractor(FieldExtractorBase):
    __slots__ = ()

    async def extract_from_form(
        self, form: FormData, *, loc: typing.Iterable[typing.Union[int, str]]
    ) -> typing.Optional[Some[typing.Any]]:
        if self.field_name not in form:
            return None
        return Some(
            await self.field_extractor.extract_from_field(
                form[self.field_name], loc=(*loc, self.field_name)
            )
        )


class RepeatedFieldExtractor(FieldExtractorBase):
    __slots__ = ()

    async def extract_from_form(
        self, form: FormData, *, loc: typing.Iterable[typing.Union[int, str]]
    ) -> typing.Optional[Some[typing.Any]]:
        return Some(
            [
                await self.field_extractor.extract_from_field(
                    val, loc=(*loc, self.field_name)
                )
                for field_name, val in form.multi_items()
                if field_name == self.field_name
            ]
        )


class FieldExtractorMarker(typing.NamedTuple):
    alias: typing.Optional[str]
    repeated: bool

    def register_parameter(self, param: inspect.Parameter) -> FormDataExtractor:
        field = model_field_from_param(param)
        field_name = self.alias or field.alias
        for marker in get_markers_from_parameter(param):
            if isinstance(marker, BodyBinderMarker):
                field_marker = marker
                break
        else:
            raise TypeError(
                "No field marker found"
                "\n You must include a valid field marker using ExtractField[AsJson[...]]"
                " or Annotated[..., Json(), Field()]"
            )
        field_extractor = field_marker.extractor_marker.register_parameter(param)
        if self.repeated:
            return RepeatedFieldExtractor(
                field_name=field_name, field_extractor=field_extractor
            )
        else:
            return FieldExtractor(
                field_name=field_name, field_extractor=field_extractor
            )



from di import BaseContainer
from starlette.responses import Response

from xpresso._utils.compat import TypedDict


class _XpressoASGIExtension(TypedDict):
    container: BaseContainer
    response_sent: bool


class XpressoASGIExtension(_XpressoASGIExtension, total=False):
    response: Response



from typing import Any, Mapping, Union

from xpresso.openapi import models as openapi_models


def parse_examples(
    examples: Mapping[str, Union[openapi_models.Example, Any]]
) -> openapi_models.Examples:
    return {
        k: v
        if isinstance(v, openapi_models.Example)
        else openapi_models.Example(value=v)
        for k, v in examples.items()
    }



from __future__ import annotations

import inspect
import typing

import di
from di.api.dependencies import DependantBase
from di.api.providers import DependencyProvider

from xpresso._utils.compat import Literal

T = typing.TypeVar("T")


Scope = Literal["app", "connection", "endpoint"]


class Depends(di.Marker, di.Dependant[typing.Any]):
    scope: Scope

    def __init__(
        self,
        call: typing.Optional[DependencyProvider] = None,
        scope: Scope = "connection",
        use_cache: bool = True,
        wire: bool = True,
        sync_to_thread: bool = False,
    ) -> None:
        super().__init__(
            call=call,
            scope=scope,
            use_cache=use_cache,
            wire=wire,
            sync_to_thread=sync_to_thread,
        )

    def from_callable(
        self, call: typing.Optional[DependencyProvider]
    ) -> DependantBase[typing.Any]:
        return Depends(
            call=call,
            scope=self.scope,
            use_cache=self.use_cache,
            wire=self.wire,
            sync_to_thread=self.sync_to_thread,
        )

    def initialize_sub_dependant(
        self, param: inspect.Parameter
    ) -> DependantBase[typing.Any]:
        if param.default is param.empty:
            # try to auto-wire
            return Depends(
                call=None,
                scope=self.scope,
                use_cache=self.use_cache,
            ).register_parameter(param)
        # has a default parameter but we create a dependency anyway just for binds
        # but do not wire it to make autowiring less brittle and less magic
        return Depends(
            call=None,
            scope=self.scope,
            use_cache=self.use_cache,
            wire=False,
        ).register_parameter(param)



import inspect
import typing

from pydantic.error_wrappers import ErrorWrapper
from starlette.datastructures import FormData

from xpresso._utils.typing import Some, model_field_from_param
from xpresso.binders._body.form_field import FormDataExtractor
from xpresso.binders._utils.forms import (
    Extractor,
    UnexpectedFileReceived,
    get_extractor,
)
from xpresso.binders.exceptions import InvalidSerialization
from xpresso.exceptions import RequestValidationError


class FormEncodedFieldExtractor:
    __slots__ = ("alias", "style", "explode", "field_name", "extractor")

    def __init__(
        self,
        alias: typing.Optional[str],
        style: str,
        explode: bool,
        field_name: str,
        extractor: Extractor,
    ) -> None:
        self.alias = alias
        self.style = style
        self.explode = explode
        self.field_name = field_name
        self.extractor = extractor

    async def extract_from_form(
        self, form: FormData, *, loc: typing.Iterable[typing.Union[int, str]]
    ) -> typing.Optional[Some[typing.Any]]:
        try:
            return self.extractor(name=self.field_name, params=form.multi_items())
        except InvalidSerialization as e:
            raise RequestValidationError(
                [
                    ErrorWrapper(
                        exc=TypeError("Data is not a valid URL encoded form"),
                        loc=tuple((*loc, self.field_name)),
                    )
                ]
            ) from e
        except UnexpectedFileReceived as exc:
            raise RequestValidationError(
                [
                    ErrorWrapper(
                        exc=exc,
                        loc=tuple((*loc, self.field_name)),
                    )
                ]
            ) from exc


class FormEncodedFieldExtractorMarker(typing.NamedTuple):
    alias: typing.Optional[str]
    style: str
    explode: bool

    def register_parameter(self, param: inspect.Parameter) -> FormDataExtractor:
        field_name = self.alias or param.name
        extractor = get_extractor(
            style=self.style, explode=self.explode, field=model_field_from_param(param)
        )
        return FormEncodedFieldExtractor(
            alias=self.alias,
            style=self.style,
            explode=self.explode,
            field_name=field_name,
            extractor=extractor,
        )
